/*
 * This file contains generic C code to print call chains, assuming
 * you have GCC >= 2. It is neither pretty nor (space and time)
 * efficient. On some architectures it could be terribly inefficient,
 * e.g. on a sparc, each expansion of __builtin_return_address()
 * might flush the register windows.
 *
 * It is therefore preferable to have an architecture specific version
 * of mprbt(). For a hint on how to do this, you could examine the
 * assembly source listing produced for function mprbt().
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <setjmp.h>

#if __GNUC__ < 2
#error you need gcc >= 2 to use this abomination
#endif

#undef ulong
#define ulong unsigned long

static ulong		lowpc, highpc;
static sigjmp_buf	jmpbuf;

extern int mprprint(int, char *, ...);

#define entrypc(pc) (pc >= lowpc && pc <= highpc)

#define printpc(n)							\
	if ((pc = (ulong)__builtin_return_address(n)) && !entrypc(pc))	\
		mprprint(fd, "%U:", pc);				\
	else								\
		goto done;

static void
mprsigsegv(int sig)
{
	siglongjmp(jmpbuf, 1);
}

void
mprbt(int fd)
{
	ulong pc;
	struct sigaction sa, osa;
	sigset_t mask, omask;

	sigemptyset(&mask);
	sigaddset(&mask, SIGSEGV);
	sigprocmask(SIG_UNBLOCK, &mask, &omask);
	sa.sa_handler = mprsigsegv;
	sa.sa_flags = 0;
	sigemptyset(&sa.sa_mask);
	sigaction(SIGSEGV, &sa, &osa);

	if (sigsetjmp(jmpbuf, 1))
		goto done;
	printpc( 0); printpc( 1); printpc( 2); printpc( 3); printpc( 4);
	printpc( 5); printpc( 6); printpc( 7); printpc( 8); printpc( 9);
	printpc(10); printpc(11); printpc(12); printpc(13); printpc(14);
	printpc(15); printpc(16); printpc(17); printpc(18); printpc(19);
	printpc(20); printpc(21); printpc(22); printpc(23); printpc(24);
	printpc(25); printpc(26); printpc(27); printpc(28); printpc(29);
	printpc(30); printpc(31); printpc(32); printpc(33); printpc(34);
	printpc(35); printpc(36); printpc(37); printpc(38); printpc(39);
done:
	sigaction(SIGSEGV, &osa, 0);
	sigprocmask(SIG_SETMASK, &omask, 0);
}

int
mprbtinit(void)
{
	char *mprpc;

	if (!(mprpc = getenv("MPRPC")))
		return 0;
	mprprint(2, "mprbtinit: warning: MPRPC is deprecated\n");
	if (!strchr(mprpc, ':')) {
		mprprint(2, "mprbtinit: error: MPRPC=%s invalid\n", mprpc);
		return -1;
	}
	lowpc  = atoi(mprpc);
	highpc = atoi(1 + strchr(mprpc, ':'));
	if (lowpc > highpc) {
		mprprint(2, "mprbtinit: error: MPRPC=%s invalid\n", mprpc);
		return -2;
	}
	return 0;
}
