#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>

#undef malloc
#undef free
#undef calloc
#undef realloc
#include <malloc.h>

#undef ulong
#define ulong unsigned long

extern void	mprbt(int);
extern int	mprbtinit(void);
extern void	*(*__mpr_malloc_hook)(size_t);
static void	*(*old_malloc_hook)(size_t);
extern void	(*__mpr_free_hook)(void *);
static void	(*old_free_hook)(void *);
extern void	*(*__mpr_realloc_hook)(void *, size_t);
static void	*(*old_realloc_hook)(void *, size_t);
static void	hook(void), unhook(void);
int		mprprint(int fd, char *fmt, ...);

static pid_t	mprmypid;	/* current process id */
static pid_t	mprpid;		/* MPRFI child process id */
static int	mprfd=-1;	/* MPRFI pipe descriptor */
static int	mprsync;	/* == 0 -> buffer writes to mprfd */
static char	mprbuf[8192];	/* buffer for mprfd */
static char	*mprbp=mprbuf;	/* current buffer pointer */

static int
xwrite(int fd, char *buf, int len)
{
	int nleft, n;

	for (nleft=len; nleft>0; nleft-=n, buf+=n) {
		n = write(fd, buf, nleft);
		if (n < 0) {
			if (errno == EINTR) {
				n=0;
				continue;
			}
			if (fd != 2)
				mprprint(2, "xwrite: warning: write failed errno=%d\n", errno);
			return len-nleft;
		}
	}
	return len;
}

static void
mprflush(void)
{
	xwrite(mprfd, mprbuf, mprbp-mprbuf);
	mprbp = mprbuf;
}

static int
mprwrite(int fd, char *str, char *buf, int len)
{
	if (str) {
		memmove(str, buf, len);
		return len;
	}
	if (fd != mprfd) {
		xwrite(fd, buf, len);
		return len;
	}
	if (mprbp+len > mprbuf+sizeof(mprbuf))	/* flush */
		mprflush();
	if (mprbp+len > mprbuf+sizeof(mprbuf))	/* can't buffer */
		xwrite(fd, buf, len);
	else {					/* buffer */
		memmove(mprbp, buf, len);
		mprbp += len;
	}
	return len;
}

static int
mprvprint(int fd, char *str, char *fmt, va_list ap)
{
	ulong ul;
	int i, len;
	char *ep, *cp, buf[64];

	for (ep=fmt, len=0; *ep; ep++) {
		if (*ep == '%') {
			switch (*++ep) {
			case 'U':
				ul = va_arg(ap, ulong);
				cp = buf + sizeof(buf);
				do
					*--cp = '0' + ul%10;
				while ((ul/=10) != 0);
				len += mprwrite(fd, str, cp, buf+sizeof(buf)-cp);
				break;
			case 'd':
				i = va_arg(ap, int);
				ul = i<0 ? -i : i;
				cp = buf + sizeof(buf);
				do
					*--cp = '0' + ul%10;
				while ((ul/=10) != 0);
				if (i < 0)
					*--cp = '-';
				len += mprwrite(fd, str, cp, buf+sizeof(buf)-cp);
				break;
			case 's':
				cp = va_arg(ap, char *);
				len += mprwrite(fd, str, cp, strlen(cp));
				break;
			default:
				abort();
			}
		}
		else {
			len += mprwrite(fd, str, ep, 1);
			if (mprsync && fd == mprfd && *ep == '\n')
				mprflush();
		}
	}
	return len;
}

int
mprprint(int fd, char *fmt, ...)
{
	int len;
	va_list ap;

	if (fd < 0)
		return -1;
	va_start(ap, fmt);
	len = mprvprint(fd, 0, fmt, ap);
	va_end(ap);
	return len;
}

static int
mprsprint(char *str, char *fmt, ...)
{
	int len;
	va_list ap;

	va_start(ap, fmt);
	len = mprvprint(-1, str, fmt, ap);
	va_end(ap);
	return len;
}

static void
mprexit(void)
{
	pid_t pid;
	int status;

	if (mprfd < 0)
		return;
	mprflush();
	if (close(mprfd) < 0)
		mprprint(2, "mprexit: warning: close failed errno=%d\n", errno);
	mprfd = -1;
	unhook();
	if (mprmypid != getpid())	/* child */
		return;
	do
		pid = waitpid(mprpid, &status, 0);
	while (pid < 0 && errno == EINTR);
	if (pid < 0) {
		mprprint(2, "mprexit: warning: waitpid failed errno=%d\n", errno);
		return;
	}
	if (WIFEXITED(status) && WEXITSTATUS(status) != 0)
		mprprint(2, "mprexit: warning: MPRFI exited %d\n", WEXITSTATUS(status));
	if (WIFSIGNALED(status))
		mprprint(2, "mprexit: warning: MPRFI caught signal %d\n", WTERMSIG(status));
}

static void *
malloc_hook(size_t sz)
{
	void *ptr;

	unhook();
	ptr = malloc(sz);
	mprprint(mprfd, "m:");
	mprbt(mprfd);
	mprprint(mprfd, "%U:%U\n", (ulong)sz, (ulong)ptr);
	hook();
	return ptr;
}

static void
free_hook(void *ptr)
{
	unhook();
	mprprint(mprfd, "f:");
	mprbt(mprfd);
	mprprint(mprfd, "%U\n", (ulong)ptr);
	free(ptr);
	hook();
}

static void *
realloc_hook(void *ptr, size_t sz)
{
	void *ptr2;

	unhook();
	ptr2 = realloc(ptr, sz);
	mprprint(mprfd, "r:");
	mprbt(mprfd);
	mprprint(mprfd, "%U:%U:%U\n", (ulong)sz, (ulong)ptr, (ulong)ptr2);
	hook();
	return ptr2;
}

static void
hook(void)
{
	old_malloc_hook = __mpr_malloc_hook;
	old_free_hook = __mpr_free_hook;
	old_realloc_hook = __mpr_realloc_hook;
	__mpr_malloc_hook = malloc_hook;
	__mpr_free_hook = free_hook;
	__mpr_realloc_hook = realloc_hook;
}

static void
unhook(void)
{
	__mpr_malloc_hook = old_malloc_hook;
	__mpr_free_hook = old_free_hook;
	__mpr_realloc_hook = old_realloc_hook;
}

static void
mprsigpipe(int sig)
{
	mprexit();
	mprprint(2, "mprsigpipe: warning: caught SIGPIPE\n");
	exit(1);
}

static int
mprpopen(char *mprfi)
{
	char cmd[256], *cp;
	int pfd[2];

	for (cp=cmd; *mprfi && cp<cmd+sizeof(cmd); ) {	/* s/%p/pid/ */
		if (mprfi[0] == '%' && mprfi[1] == 'p') {
			cp += mprsprint(cp, "%d", getpid());
			mprfi += 2;
		}
		else
			*cp++ = *mprfi++;
	}
	*cp = '\0';

	if (pipe(pfd) < 0)
		return -1;
	if ((mprpid = fork()) < 0) {
		close(pfd[0]);
		close(pfd[1]);
		return -1;
	}
	else if (mprpid == 0) {
		close(pfd[1]);
		if (pfd[0] != 0) {
			dup2(pfd[0], 0);
			close(pfd[0]);
		}
		putenv("MPRFI");
		execl("/bin/sh", "sh", "-c", cmd, 0);
		mprprint(2, "mprpopen: error: could not exec /bin/sh\n");
		_exit(127);
	}
	close(pfd[0]);
	return pfd[1];
}

int
mpr(void)
{
	char *mprfi;
	static int mprinit;

	if (mprinit)
		return 0;
	mprmypid = getpid();
	mprinit = 1;
	if (!(mprfi = getenv("MPRFI")) || !*mprfi)
		return 0;
	if (mprbtinit() < 0)
		return -1;
	if ((mprfd = mprpopen(mprfi)) < 0) {
		mprprint(2, "mpr: error: mprpopen() failed\n");
		return -2;
	}
	if (atexit(mprexit) < 0) {
		mprprint(2, "mpr: error: atexit failed\n");
		close(mprfd);
		return -3;
	}
#ifdef MPRSIGPIPE
	signal(SIGPIPE, mprsigpipe);
#endif
	if (getenv("MPRSYNC"))
		mprsync = 1;
	if (getenv("MPRNOCHILD"))
		putenv("MPRFI");
	hook();
	return 0;
}
